/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/* this cell renderer is based on GConfCellRenderer from gconf-editor,
   Copyright (C) 2001, 2002 Anders Carlsson <andersca@gnu.org> */

#include "configure_cell_renderer.h"
#include "gpt_marshal.h"

enum {
  PROP_VALUE = 1,
  PROP_TYPE,
  PROP_MODIFIED,
};

enum {
  CHANGED,
  CHECK_WRITABLE,
  LAST_SIGNAL,
};

static guint configure_cell_signals[LAST_SIGNAL] = { 0 };

#define CONFIGURE_CELL_RENDERER_TEXT_PATH "configure-cell-renderer-text-path"
#define CONFIGURE_CELL_RENDERER_VALUE "configure-cell-renderer-value"
#define CONFIGURE_CELL_RENDERER_VALUE_TYPE "configure-cell-renderer-value-type"

static void
configure_cell_renderer_get_property (GObject    *object,
                                      guint      param_id,
                                      GValue     *value,
                                      GParamSpec *pspec)
{
  ConfigureCellRenderer *cell;
  cell = CONFIGURE_CELL_RENDERER (object);
	
  switch (param_id)
    {
      case PROP_VALUE:
        g_value_set_string (value, cell->value);
        break;
      
      case PROP_TYPE:
        g_value_set_int (value, cell->type);
        break;
      
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
        break;
    }
}

static void
configure_cell_renderer_set_property (GObject      *object,
                                      guint        param_id,
                                      const GValue *value,
                                      GParamSpec   *pspec)
{
  ConfigureCellRenderer *cell;
  gchar *cell_value;
  GType cell_type;
  GtkCellRendererMode new_mode = GTK_CELL_RENDERER_MODE_INERT;
	
  cell = CONFIGURE_CELL_RENDERER (object);

  switch (param_id)
    {
      case PROP_VALUE:
        if (cell->value) 
          g_free (cell->value);
        cell_value = (gchar *) g_value_get_string (value);

        if (cell_value)
            cell->value = g_strdup (cell_value);
        else
            cell->value = g_strdup ("");
          
        break;
          
      case PROP_TYPE:
        cell_type = g_value_get_int (value);
      
        switch (cell_type)
          {
            case G_TYPE_STRING:
              new_mode = GTK_CELL_RENDERER_MODE_EDITABLE;
              break;
				
            case G_TYPE_BOOLEAN:
              new_mode = GTK_CELL_RENDERER_MODE_ACTIVATABLE;
              break;

            default:
              g_warning ("unhandled value type %d", (int) cell_type);
              break;
          }

        cell->type = cell_type;
        g_object_set (object, "mode", new_mode, NULL);
        break;
          
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
        break;	
    }
}

static void
configure_cell_renderer_text_editing_done (GtkCellEditable        *entry,
                                           ConfigureCellRenderer  *cell)
{
  const gchar *path;
  const gchar *new_text;
  const gchar *value;
  int type;
	
  if (GTK_ENTRY (entry)->editing_canceled)
      return;

  path = g_object_get_data (G_OBJECT (entry),
                            CONFIGURE_CELL_RENDERER_TEXT_PATH);
  value = g_object_get_data (G_OBJECT (entry),
                             CONFIGURE_CELL_RENDERER_VALUE);
  type = (gint) g_object_get_data (G_OBJECT (entry),
                                   CONFIGURE_CELL_RENDERER_VALUE_TYPE);
  new_text = gtk_entry_get_text (GTK_ENTRY (entry));

  switch (type)
    {
      case G_TYPE_STRING:
        value = new_text;
        break;
      default:
        g_error ("editing done, unknown value %d", type);
        break;
    }
	
  g_signal_emit (cell,
                 configure_cell_signals[CHANGED],
                 0, path, value);
}

static gboolean
configure_cell_renderer_check_writability (ConfigureCellRenderer *cell,
                                           const gchar           *path)
{
  gboolean ret = TRUE;
  g_signal_emit (cell,
                 configure_cell_signals[CHECK_WRITABLE],
                 0, path, &ret);
  return ret;
}

GtkCellEditable *
configure_cell_renderer_start_editing (GtkCellRenderer      *cell,
                                       GdkEvent             *event,
                                       GtkWidget            *widget,
                                       const gchar          *path,
                                       GdkRectangle         *background_area,
                                       GdkRectangle         *cell_area,
                                       GtkCellRendererState  flags)
{
  GtkWidget *entry;
  ConfigureCellRenderer *cellvalue;

  cellvalue = CONFIGURE_CELL_RENDERER (cell);

  /* If not writable then we definately can't edit */
  if ( ! configure_cell_renderer_check_writability (cellvalue, path))
      return NULL;
	
  switch (cellvalue->type)
    {
      case G_TYPE_STRING:
        entry = g_object_new (GTK_TYPE_ENTRY,
                              "has_frame", FALSE,
                              "text", cellvalue->value,
                              NULL);
        g_signal_connect (entry, "editing_done",
                          G_CALLBACK (configure_cell_renderer_text_editing_done),
                          cellvalue);

        g_object_set_data_full (G_OBJECT (entry),
                                CONFIGURE_CELL_RENDERER_TEXT_PATH,
                                g_strdup (path),
                                g_free);
        g_object_set_data_full (G_OBJECT (entry),
                                CONFIGURE_CELL_RENDERER_VALUE,
                                g_strdup (cellvalue->value),
                                g_free);
        g_object_set_data (G_OBJECT (entry),
                           CONFIGURE_CELL_RENDERER_VALUE_TYPE,
                           (gpointer) cellvalue->type);
		
        gtk_widget_show (entry);

        return GTK_CELL_EDITABLE (entry);
        break;
        
      default:
        g_error ("%d shouldn't be handled here", (int) cellvalue->type);
        break;
    }
    
  return NULL;
}

static gint
configure_cell_renderer_activate (GtkCellRenderer *cell,
                                  GdkEvent        *event,
                                  GtkWidget       *widget,
                                  const gchar     *path,
                                  GdkRectangle    *background_area,
                                  GdkRectangle    *cell_area,
                                  guint            flags)
{
  ConfigureCellRenderer *cellvalue;
  cellvalue = CONFIGURE_CELL_RENDERER (cell);

  if (cellvalue->value == NULL) /* FIXME? */
      return TRUE;

  switch (cellvalue->type)
    {
      case G_TYPE_BOOLEAN:
        if (cellvalue->value[0] != '\0')
            cellvalue->value[0] = '\0';
        else
            cellvalue->value[0] = '1';
        
        g_signal_emit (cell,
                       configure_cell_signals[CHANGED],
                       0, path, cellvalue->value);
		
        break;
      
      default:
        g_error ("%d shouldn't be handled here", (int) cellvalue->type);
        break;
    }

  return TRUE;
}

static void
configure_cell_renderer_get_size (GtkCellRenderer *cell,
                                  GtkWidget       *widget,
                                  GdkRectangle    *cell_area,
                                  gint            *x_offset,
                                  gint            *y_offset,
                                  gint            *width,
                                  gint            *height)
{
  ConfigureCellRenderer *cellvalue;
	
  cellvalue = CONFIGURE_CELL_RENDERER (cell);

/*  if (cellvalue->value == NULL) {
		g_object_set (G_OBJECT (cellvalue->text_renderer),
			      "text", "<no value>",
			      NULL);
		gtk_cell_renderer_get_size (cellvalue->text_renderer, widget, cell_area,
					    x_offset, y_offset, width, height);
		return;
} */
	
  switch (cellvalue->type)
    {
      case G_TYPE_STRING:
        g_object_set (G_OBJECT (cellvalue->text_renderer),
                      "text", cellvalue->value,
                      NULL);
        gtk_cell_renderer_get_size (cellvalue->text_renderer,
                                    widget, cell_area,
                                    x_offset, y_offset,
                                    width, height);
        break;
      
      case G_TYPE_BOOLEAN:
        g_object_set (G_OBJECT (cellvalue->toggle_renderer),
                      "xalign", 0.0,
                      "active", ((cellvalue->value)[0] != '\0'),
                      NULL);
        gtk_cell_renderer_get_size (cellvalue->toggle_renderer,
                                    widget, cell_area,
                                    x_offset, y_offset,
                                    width, height);
                                    break;

      default:
        g_warning ("get_size: Unknown type: %d\n", (int) cellvalue->type);
        break;
    }
}

static void
configure_cell_renderer_render (GtkCellRenderer         *cell,
                                GdkWindow               *window,
                                GtkWidget               *widget,
                                GdkRectangle            *background_area,
                                GdkRectangle            *cell_area,
                                GdkRectangle            *expose_area,
                                GtkCellRendererState    flags)
{
  ConfigureCellRenderer *cellvalue;
	
  cellvalue = CONFIGURE_CELL_RENDERER (cell);

/*  if (cellvalue->value == NULL) {
		g_object_set (G_OBJECT (cellvalue->text_renderer),
			      "text", "<no value>",
			      NULL);

		gtk_cell_renderer_render (cellvalue->text_renderer, window, widget,
					  background_area, cell_area, expose_area, flags);
		return;
	} */

  switch (cellvalue->type)
    {
      case G_TYPE_STRING:
        g_object_set (G_OBJECT (cellvalue->text_renderer),
                      "text", cellvalue->value,
                      NULL);
        gtk_cell_renderer_render (cellvalue->text_renderer,
                                  window, widget,
                                  background_area, cell_area,
                                  expose_area, flags);
        break;
    
      case G_TYPE_BOOLEAN:
        g_object_set (G_OBJECT (cellvalue->toggle_renderer),
                      "xalign", 0.0,
                      "active", ((cellvalue->value)[0] != '\0'),
                      NULL);
		
        gtk_cell_renderer_render (cellvalue->toggle_renderer,
                                  window, widget,
                                  background_area, cell_area,
                                  expose_area, flags);
                                  break;
      default:
        g_warning ("render: Unknown type: %d\n", (int) cellvalue->type);
        break;
    }
}

static void
configure_cell_renderer_class_init (ConfigureCellRendererClass *klass)
{
  GObjectClass *object_class = (GObjectClass *)klass;
  GtkCellRendererClass *cell_renderer_class = (GtkCellRendererClass *)klass;
	
  object_class->get_property = configure_cell_renderer_get_property;
  object_class->set_property = configure_cell_renderer_set_property;

  cell_renderer_class->get_size = configure_cell_renderer_get_size;
  cell_renderer_class->render = configure_cell_renderer_render;
  cell_renderer_class->activate = configure_cell_renderer_activate;
  cell_renderer_class->start_editing = configure_cell_renderer_start_editing;

  g_object_class_install_property (object_class, PROP_VALUE,
                                   g_param_spec_string ("value",
                                                        "Value",
                                                        "Value",
                                                        "",
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_TYPE,
                                   g_param_spec_int    ("type",
                                                        "Type",
                                                        "Type",
                                                        0, 65535, 0,
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_MODIFIED,
                                   g_param_spec_boolean("modified",
                                                        "Modified",
                                                        "Modified",
                                                        FALSE,
                                                        G_PARAM_READWRITE));
                                                        
  configure_cell_signals[CHANGED] =
              g_signal_new ("changed",
                            G_TYPE_FROM_CLASS (object_class),
                            G_SIGNAL_RUN_LAST,
                            G_STRUCT_OFFSET (ConfigureCellRendererClass,
                                             changed),
                            (GSignalAccumulator) NULL, NULL,
                            gpt_marshal_VOID__STRING_STRING,
			    G_TYPE_NONE, 2,
			    G_TYPE_STRING,
			    G_TYPE_STRING);
  configure_cell_signals[CHECK_WRITABLE] =
		g_signal_new ("check_writable",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (ConfigureCellRendererClass, changed),
			      (GSignalAccumulator) NULL, NULL,
			      gpt_marshal_BOOLEAN__STRING,
			      G_TYPE_BOOLEAN, 1,
			      G_TYPE_STRING);
}

static void
configure_cell_renderer_init (ConfigureCellRenderer *renderer)
{
  GTK_CELL_RENDERER (renderer)->mode = GTK_CELL_RENDERER_MODE_ACTIVATABLE;
	
  renderer->text_renderer = gtk_cell_renderer_text_new ();
  renderer->toggle_renderer = gtk_cell_renderer_toggle_new ();
}

GType
configure_cell_renderer_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
        {
          sizeof (ConfigureCellRendererClass),
          NULL,		/* base_init */
          NULL,		/* base_finalize */
          (GClassInitFunc) configure_cell_renderer_class_init,
          NULL,		/* class_finalize */
          NULL,		/* class_data */
          sizeof (ConfigureCellRenderer),
          0,              /* n_preallocs */
          (GInstanceInitFunc) configure_cell_renderer_init
        };
		
      object_type = g_type_register_static (GTK_TYPE_CELL_RENDERER,
                                            "ConfigureCellRenderer",
                                            &object_info, 0);
    }

  return object_type;
}

GtkCellRenderer *
configure_cell_renderer_new (void)
{
  return GTK_CELL_RENDERER (g_object_new (CONFIGURE_CELL_RENDERER_TYPE,
                                          NULL));
}
