#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib.h>

#include "misc.h"
#include "pkglist.h"
#include "packages.h"
#include "treeview.h"

static void pkg_list_class_init (PkgListClass *klass);
static void pkg_list_init (PkgList *pl);
static void pkg_list_finalize (GObject *object);
static gboolean pkg_list_search_func (GtkTreeModel *model, gint column, const gchar *key, GtkTreeIter *iter, gpointer search_data);

static void pixbuf_data_func (GtkTreeViewColumn *tree_column, GtkCellRenderer *cell, GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data);
static void details_data_func (GtkTreeViewColumn *tree_column, GtkCellRenderer *cell, GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data);
static int compare_package_name (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data);

/* static void pkg_list_set_property (GObject *obj, guint prop_id, const GValue *val, GParamSpec *pspec);
static void pkg_list_get_property (GObject *obj, guint prop_id, GValue *val, GParamSpec *pspec); */

static GtkTreeViewClass *parent_class = NULL;

GType
pkg_list_get_type (void)
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (PkgListClass),
			NULL,   /* base_init */
			NULL,   /* base_finalize */
			(GClassInitFunc) pkg_list_class_init,   /* class_init */
			NULL,   /* class_finalize */
			NULL,   /* class_data */
			sizeof (PkgList),
			0,      /* n_preallocs */
			(GInstanceInitFunc) pkg_list_init    /* instance_init */
		};
		
		type = g_type_register_static (GTK_TYPE_TREE_VIEW, "PkgList", &info, 0);
	}
	return type;
}

static void
pkg_list_class_init (PkgListClass *klass)
{
	GObjectClass *object_class;

	parent_class = (GtkTreeViewClass*) g_type_class_peek_parent (klass);
	object_class = (GObjectClass*) klass;

	object_class->finalize = pkg_list_finalize;
	
	/* object_class->set_property = pkg_list_set_property;
	object_class->get_property = pkg_list_get_property; */
}

static void
pkg_list_init (PkgList *pl)
{
	g_assert (pl != NULL);
	g_assert (IS_PKG_LIST (pl));
	
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkTreeSelection *selection;
  
	pl->store = gtk_list_store_new (1, PKG_ENTRY_TYPE);
	gtk_tree_view_set_model (GTK_TREE_VIEW (pl), GTK_TREE_MODEL (pl->store));

	/* 1st column: package pixmap */
	column = gtk_tree_view_column_new ();
	renderer = gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_column_pack_start (column, renderer, TRUE);
	gtk_tree_view_column_set_cell_data_func (column, renderer, pixbuf_data_func, NULL, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (pl), column);
	
	/* 2nd column: package details */
	column = gtk_tree_view_column_new ();
	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_column_pack_start (column, renderer, TRUE);
	gtk_tree_view_column_set_cell_data_func (column, renderer, details_data_func, NULL, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (pl), column);
	
	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (pl->store), SORT_NAME, compare_package_name, NULL, NULL);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(pl->store), SORT_NAME, GTK_SORT_ASCENDING);
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (pl));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
	
	gtk_tree_view_set_enable_search (GTK_TREE_VIEW (pl), TRUE);
	gtk_tree_view_set_search_column (GTK_TREE_VIEW (pl), 0);
	gtk_tree_view_set_search_equal_func (GTK_TREE_VIEW (pl), pkg_list_search_func, NULL, NULL);
}

static void
pkg_list_finalize (GObject *object)
{
	(G_OBJECT_CLASS (parent_class)->finalize) (object);
}

GtkWidget *
pkg_list_new ()
{
	return GTK_WIDGET (g_object_new (PKG_LIST_TYPE, NULL));
}

static void
pixbuf_data_func (GtkTreeViewColumn *tree_column,
                  GtkCellRenderer   *cell,
                  GtkTreeModel      *tree_model,
                  GtkTreeIter       *iter,
                  gpointer           data)
{
	PkgEntry *entry;

	gtk_tree_model_get (tree_model, iter, 0, &entry, -1);
	g_object_set (cell, "pixbuf", entry->pixbuf, NULL);
	g_object_set (cell, "xalign", 1.0, NULL);
	g_object_set (cell, "yalign", 0.0, NULL);
	g_object_set (cell, "xpad", 2, NULL);
	g_object_set (cell, "ypad", 2, NULL);
	g_object_unref (entry);
}

static void
details_data_func (GtkTreeViewColumn *tree_column,
                  GtkCellRenderer   *cell,
                  GtkTreeModel      *tree_model,
                  GtkTreeIter       *iter,
                  gpointer           data)
{
	PkgEntry *entry;
	gchar *str;
	gchar *name, *version, *summary;
	glong size;
	
	gtk_tree_model_get (tree_model, iter, 0, &entry, -1);
	g_object_get (entry,
					"name", &name,
					"version", &version,
					"size", &size,
					"summary", &summary,
					NULL);
	str = g_strdup_printf("<big><b>%s</b></big>\nversion <b>%s</b> (%s)\n%s",
	                             name, version, parse_package_size(size), summary);
	g_object_set (cell, "markup", str, NULL);
	g_object_unref (entry);
	g_free (str);
	g_free (name);
	g_free (version);
	g_free (summary);
}

static int
compare_package_name (GtkTreeModel *model,
                      GtkTreeIter  *a,
                      GtkTreeIter  *b,
                      gpointer      user_data)
{
	PkgEntry *entry_a;
	PkgEntry *entry_b;
	
	int result;

	gtk_tree_model_get (model, a, 0, &entry_a, -1);
	gtk_tree_model_get (model, b, 0, &entry_b, -1);

	result = strcasecmp (entry_a->name, entry_b->name);

	g_object_unref (entry_a);
	g_object_unref (entry_b);

	return result;
}

static gboolean
pkg_list_search_func (GtkTreeModel  *model,
                      gint          column,
                      const gchar   *key,
                      GtkTreeIter   *iter,
                      gpointer      search_data)
{
  PkgEntry *entry;
  gtk_tree_model_get (model, iter, 0, &entry, -1);

  if (entry)
    {
      if (strlen (key) > strlen (entry->name))
        {
          g_object_unref (entry);
          return TRUE;
        }
      else if (strncmp (entry->name, key, strlen (key)) == 0)
        {
          g_object_unref (entry);
          return FALSE;
        }
    }
  
  return TRUE;
}

gboolean
packages_list_load (PkgList     *pl,
                    GError      **error)
{
  GtkTreeIter iter;
  PkgEntry *entry;
  DIR *pkgdir;

  pkgdir = opendir(PACKAGES_DIR);
  if(pkgdir == NULL )
    {
      switch (errno)
        {
          case EACCES:
            g_error ("Cannot read packages list: Access denied.");
            break;
          case ENOMEM:
            g_error ("Cannot read packages list: Not enough memory.");
            break;
          default:
            g_error ("Cannot read packages list.");
            break;
        }
      return FALSE;
    }
	
  gtk_list_store_clear (pl->store);

  struct dirent *d;
          
  while((d = readdir(pkgdir)) != NULL)
    {
      if(d->d_name[0] != '.')
        {
          gchar *uri = g_strdup_printf ("%s/%s", PACKAGES_DIR, d->d_name);
          entry = pkg_entry_new(uri);
          g_assert (entry != NULL);
          
          if(entry)
            {
              gtk_list_store_append (pl->store, &iter);
              gtk_list_store_set (pl->store, &iter, 0, entry, -1);
              g_object_unref (entry);
            }
          g_free (uri);
        }
    }
  closedir(pkgdir);
      
  if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pl->store), &iter))
    {
      GtkTreeSelection *selection;
      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (pl));
      gtk_tree_selection_select_iter (selection, &iter);
    }
  return TRUE;
}

gboolean
packages_list_refresh (PkgList  *pl,
                       GError   **error)
{
  GtkTreeIter iter, new_iter;
  PkgEntry *entry;
  struct stat *ss;
  gchar *uri;
  
  if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pl->store),
                                     &iter))
    {
      ss = g_new0 (struct stat, 1);

      GHashTable *hash_table;
      hash_table = g_hash_table_new_full (g_str_hash,
                                          g_str_equal,
                                          g_free,
                                          g_free);

      do 
        {
          gtk_tree_model_get (GTK_TREE_MODEL (pl->store),
                              &iter,
                              0, &entry,
                              -1);
        
          if (stat (entry->uri, ss) == -1)
            {
              if (errno == ENOENT) /* the file has been deleted */
                {
                  new_iter = iter;
                  tree_model_iter_prev (GTK_TREE_MODEL (pl->store), &iter);
                  gtk_list_store_remove (pl->store, &new_iter);
                }
              else
                  return FALSE;
            }
          else 
            {
              g_hash_table_insert (hash_table,
                                   g_strdup (entry->uri),
                                   g_strdup (entry->uri));
            
              if ( (entry->mtime < ss->st_mtime) ||
                   (entry->mtime <= 1) )        /* the file has been modified */
                {
                  uri = g_strdup (entry->uri);
                  new_iter = iter;
                  tree_model_iter_prev (GTK_TREE_MODEL (pl->store), &iter);
                  gtk_list_store_remove (pl->store, &new_iter);
                
                  PkgEntry *new_entry;
                  new_entry = pkg_entry_new (uri);
                  g_assert (new_entry != NULL);
                  
                  gtk_list_store_append (pl->store, &new_iter);
                  gtk_list_store_set (pl->store, &new_iter, 0, new_entry, -1);
                  g_object_unref (new_entry);

                  g_free (uri);
                }
            }
          
          g_object_unref (entry);
          
        } while(gtk_tree_model_iter_next (GTK_TREE_MODEL (pl->store),
                                          &iter));
      g_free (ss);
      
      /* now look for new files in /var/log/packages */
      DIR *pkgdir = opendir(PACKAGES_DIR);
      if(pkgdir == NULL)
        {
          switch (errno)
            {
              case EACCES:
                g_error ("Cannot read packages list: Access denied.");
                break;
              case ENOMEM:
                g_error ("Cannot read packages list: Not enough memory.");
                break;
              default:
                g_error ("Cannot read packages list.");
                break;
            }
          return FALSE;
        }
	
      struct dirent *d;
      gchar *uri;
          
      while ((d = readdir (pkgdir)) != NULL)
        {
          if(d->d_name[0] != '.')
            {
              uri = g_strdup_printf ("%s/%s", PACKAGES_DIR, d->d_name);
              if (g_hash_table_lookup (hash_table, uri) == NULL)
                {
                  entry = pkg_entry_new(uri);
                  g_assert (entry != NULL);
          
                  if(entry)
                    {
                      gtk_list_store_append (pl->store, &iter);
                      gtk_list_store_set (pl->store, &iter, 0, entry, -1);
                      g_object_unref (entry);
                    }
                }
              g_free (uri);
            }
        }
      closedir(pkgdir);
      
      g_hash_table_destroy (hash_table);
      
      return TRUE;
    }
  
  else
    {
      g_warning ("Treeview is empty?");
      return FALSE;
    }
}

gboolean
packages_list_reload (PkgList *pl, GError **error)
{
  return packages_list_refresh (pl, error);
}

gboolean
packages_list_find (PkgList *pl, const gchar *pkg_name) {
  GtkTreeIter iter;
  PkgEntry *entry;
	
  if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pl->store), &iter))
    {
      gtk_tree_model_get (GTK_TREE_MODEL (pl->store), &iter,
                          0, &entry, -1);
      if (strcmp (entry->name, pkg_name) == 0)
          return TRUE;

      while (gtk_tree_model_iter_next (GTK_TREE_MODEL (pl->store), &iter))
        {
          gtk_tree_model_get (GTK_TREE_MODEL (pl->store), &iter, 0, &entry, -1);
          if (strcmp (entry->name, pkg_name) == 0)
            {
              g_object_unref (entry);
              return TRUE;
            }
          g_object_unref (entry);
        }
    }
  
  return FALSE;
}
/* end */
